from sdc11073 import namespaces
from sdc11073 import nomenclature
from sdc11073 import pmtypes
from sdc11073.roles import providerbase
from sdc11073.sdcdevice import sco


class OperationProvider(providerbase.ProviderRole):
    """Handles operations that work on operation states.
     Empty implementation, not needed/used for pysdc tests"""
    pass
